import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable
import torch.optim as optim
import seaborn as sns
from torch import linalg 
import math
from torch.nn.parameter import Parameter
from layers import Linear, Conv2d

class MLP(nn.Module):
    
    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: torch.Tensor

    def __init__(self, in_features: int, out_features: int,
                 device=None, dtype=None) -> None:
        super(MLP, self).__init__()
        factory_kwargs = {'device': device, 'dtype': dtype}
        
        self.linear1=Linear(in_features,256)
        self.linear2=Linear(256,64)
        self.linear3=Linear(64,out_features)
       
    
    def fill_gd(self,alpha):
        self.linear1.update_grad(alpha)
        self.linear2.update_grad(alpha)
        self.linear3.update_grad(alpha)
        
    def kill_gd(self):
        self.linear1.reset_grad()
        self.linear2.reset_grad()
        self.linear3.reset_grad()
        
#     def update_gd(self,alpha):
#         self.linear1.weight.data += -alpha*self.linear1.weight.grad
#         self.linear2.weight.data += -alpha*self.linear2.weight.grad
#         self.linear3.weight.data += -alpha*self.linear3.weight.grad
        
    def forward(self, input: torch.Tensor) -> torch.Tensor:

        x = F.relu(self.linear1(input))
        x = F.relu(self.linear2(x))
        x = self.linear3(x)

        return x
    
from collections import OrderedDict
import torch.nn as nn


class SmallCNN(nn.Module):
    def __init__(self, drop=0.5):
        super(SmallCNN, self).__init__()

        self.num_channels = 1
        self.num_labels = 10

        activ = nn.ReLU(True)

        self.feature_extractor = nn.Sequential(OrderedDict([
            ('conv1', Conv2d(self.num_channels, 32, 3)),
            ('relu1', activ),
            ('conv2', Conv2d(32, 32, 3)),
            ('relu2', activ),
            ('maxpool1', nn.MaxPool2d(2, 2)),
            ('conv3', Conv2d(32, 64, 3)),
            ('relu3', activ),
            ('conv4', Conv2d(64, 64, 3)),
            ('relu4', activ),
            ('maxpool2', nn.MaxPool2d(2, 2)),
        ]))

        self.classifier = nn.Sequential(OrderedDict([
            ('fc1', Linear(64 * 4 * 4, 200)),
            ('relu1', activ),
            ('drop', nn.Dropout(drop)),
            ('fc2', Linear(200, 200)),
            ('relu2', activ),
            ('fc3', Linear(200, self.num_labels)),
        ]))

        for m in self.modules():
            if isinstance(m, (nn.Conv2d)):
                nn.init.kaiming_normal_(m.weight)
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
        nn.init.constant_(self.classifier.fc3.weight, 0)
        nn.init.constant_(self.classifier.fc3.bias, 0)

    def forward(self, input):
        features = self.feature_extractor(input)
        logits = self.classifier(features.view(-1, 64 * 4 * 4))
        return logits
    
    def fill_gd(self,alpha):
        for m in self.modules():
            if isinstance(m, (Conv2d)) or isinstance(m, (Linear)):
                m.update_grad(alpha)
#         self.linear1.
#         self.linear2.update_grad(alpha)
#         self.linear3.update_grad(alpha)
        
    def kill_gd(self):
        for m in self.modules():
            if isinstance(m, (Conv2d)) or isinstance(m, (Linear)):
                m.reset_grad()
#         self.linear1.reset_grad()
#         self.linear2.reset_grad()
#         self.linear3.reset_grad()