from train_utils import train_robust_data, test, train, attack
from resnet import ResNet18
import torch
import torchvision.transforms as T
import torchvision.datasets as datasets
from torch.utils.data import TensorDataset
import os
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
import numpy as np
import argparse
import pickle
from core import Smooth
import time
import datetime
from tqdm import tqdm


np.random.seed(42)
# Optimization Criterion
loss_name = 'CrossEntropyLoss'
# loss_name = 'MSELoss'

# Optimization hyperparameters
lr_decay = 0.01

C = 10

transform = T.Compose([T.ToTensor()])


train_data = datasets.CIFAR10('../data', train=True, download=True, transform=transform)
val_data = datasets.CIFAR10('../data', train=False, download=True, transform=transform)


def shuffle_datasets(d1, d2):
    assert len(d1) == len(d2)
    indices = np.arange(len(d1))
    np.random.shuffle(indices)
    return torch.utils.data.Subset(d1, indices), torch.utils.data.Subset(d2, indices)


def robust_test(model, device, robust_train_loader, test_loader, criterion, num_classes, optimizer, epoch, noise_sd=None):
    train(model, criterion, device, num_classes, robust_train_loader, optimizer, epoch, noise_sd)


def run():
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model = ResNet18().to(device)
    model.load_state_dict(torch.load(args.model_ckpt, map_location=device))

    if loss_name == 'CrossEntropyLoss':
        criterion = nn.CrossEntropyLoss()
        criterion_summed = nn.CrossEntropyLoss(reduction='sum')

    optimizer = optim.SGD(model.parameters(),
                          lr=lr,
                          momentum=momentum,
                          weight_decay=weight_decay
                          )

    lr_scheduler = optim.lr_scheduler.MultiStepLR(optimizer,
                                                  milestones=epochs_lr_decay,
                                                  gamma=lr_decay)

    val_loader = torch.utils.data.DataLoader(val_data, batch_size=batch_size, shuffle=False)
    if args.load:
        f_name = f'saved_datasets/{file_prefix}.pickle'
        print(f'Reading file: {f_name}')
        with open(f_name, 'rb') as handle:
            dataset = pickle.load(handle)
    else:
        dataset = {}
        for epoch in range(1, 12):
            if epoch == 1:
                train_dataset, robust_train_dataset = shuffle_datasets(train_data, train_data)
                train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
                robust_train_loader = torch.utils.data.DataLoader(robust_train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
            if epoch > 1:
                robust_train_dataset = TensorDataset(robust_dataset, robust_label)
                train_dataset, robust_train_dataset = shuffle_datasets(train_dataset, robust_train_dataset)
                train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
                robust_train_loader = torch.utils.data.DataLoader(robust_train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
            if epoch % 1 == 0:
                dataset[str(epoch)] = robust_train_dataset
            robust_dataset, robust_label = train_robust_data(model,
                                                             criterion, device, train_loader, robust_train_loader,
                                                             optimizer, epoch, attack_type=args.type, alpha0=alpha0, beta=beta, eps=eps,
                                                             step_size=alpha, steps=steps, noise_sd=args.noise_sd)
            test(model, device, val_loader, noise_sd=args.noise_sd)

        # save the dataset:

        with open(f'saved_datasets/{file_prefix}.pickle', 'wb') as handle:
            pickle.dump(dataset, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # robust_train_loader = torch.utils.data.DataLoader(dataset["12"],batch_size=batch_size, shuffle=False)

    print('Fine tuning:')
    for key in dataset:
        if int(key) % 2 != 0:
            continue
        print("************", key, "*********************")
        robust_train_loader = torch.utils.data.DataLoader(dataset[key], batch_size=batch_size, shuffle=False)

        model22 = ResNet18().to(device)
        model22.load_state_dict(torch.load(args.model_ckpt, map_location=device))
        optimizer = optim.SGD(model22.parameters(),
                              lr=lr,
                              momentum=momentum,
                              weight_decay=weight_decay
                              )

        for i in range(21):
            robust_test(model22, device, robust_train_loader, val_loader, criterion, 200, optimizer, i, args.noise_sd)
            if i % 3 == 0:
                smoothed_f = Smooth(model22, C, args.noise_sd)

                # prepare output file
                f = open(f'{args.outfile}_key{key}_i{i}', 'w')
                print("idx\tlabel\tpredict\tradius\tcorrect\ttime", file=f, flush=True)

                for idx in tqdm(range(len(val_data))):
                    (x, label) = val_data[idx]

                    before_time = time.time()
                    # certify the prediction of g around x
                    x = x.cuda()
                    prediction, radius = smoothed_f.certify(x, args.N0, args.N, args.prob, 10000)
                    after_time = time.time()
                    correct = int(prediction == label)

                    time_elapsed = str(datetime.timedelta(seconds=(after_time - before_time)))
                    print("{}\t{}\t{}\t{:.3}\t{}\t{}".format(
                        idx, label, prediction, radius, correct, time_elapsed), file=f, flush=True)

                f.close()


def args_parser():
    parser = argparse.ArgumentParser()

    # command line arguments
    parser.add_argument('--lr', type=float, default=None)
    #parser.add_argument('--epochs', type=int, default=None)
    parser.add_argument('--bs', type=int, default=None)
    parser.add_argument('--alpha', type=float, default=None)
    parser.add_argument('--eps', type=float, default=None)
    parser.add_argument('--beta', type=float, default=None)
    parser.add_argument('--alpha0', type=float, default=None)
    parser.add_argument('--type', type=str, default='l2')
    parser.add_argument('--model_ckpt', type=str, default=None)

    parser.add_argument('--noise_sd', type=float, default=None)
    parser.add_argument('--N0', type=int, default=None)
    parser.add_argument('--N', type=int, default=None)
    parser.add_argument('--prob', type=float, default=None)
    parser.add_argument('--outfile', type=str, default=None)

    parser.add_argument('--load', type=int, default=0)
    parser.add_argument('--noise_free', type=int, default=0)
    return parser.parse_args()


if __name__ == "__main__":
    # Best lr after hyperparameter tuning

    args = args_parser()

    lr = 0.01 if args.lr is None else args.lr

    epochs = 100
    epochs_lr_decay = [epochs / 3, epochs * 2 / 3]


    momentum = 0.9
    weight_decay = 5e-4

    batch_size = 1000 if args.bs is None else args.bs
    eps = 0.25 if args.eps is None else args.eps
    alpha = 0.06 if args.alpha is None else args.alpha
    steps = 10
    beta = 1 / 255 if args.beta is None else args.beta
    alpha0 = 0.01 if args.alpha0 is None else args.alpha0

    if args.noise_free:
        file_prefix = f'rs_lr{lr}_bs{batch_size}_eps{eps}_beta{beta}_alpha0{alpha0}_alpha{alpha}_noise_free'
    else:
        file_prefix = f'rs_lr{lr}_bs{batch_size}_eps{eps}_beta{beta}_alpha0{alpha0}_alpha{alpha}_sd{args.noise_sd}'
    print(file_prefix)
    run()
