from train_utils import train_robust_data, test, train, attack
from resnet_imgnet import resnet18
import torch
import torchvision.transforms as T
import torchvision.datasets as datasets
from torch.utils.data import TensorDataset
import os
import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler
import numpy as np
import argparse
import pickle

np.random.seed(42)
# Optimization Criterion
loss_name = 'CrossEntropyLoss'
# loss_name = 'MSELoss'

# Optimization hyperparameters
lr_decay = 0.1


# analysis parameters
# epoch_list          = [1,   2,   3,   4,   5,   6,   7,   8,   9,   10,   11,
#                        12,  13,  14,  16,  17,  19,  20,  22,  24,  27,   29,
#                        32,  35,  38,  42,  45,  50,  54,  59,  65,  71,   77,
#                        85,  92,  101, 110, 121, 132, 144, 158, 172, 188,  206,
#                        225, 245, 268, 293, 320, 350]

epoch_list = [1, 3, 5, 7, 11,
              12, 17, 22, 27, 29,
              32, 38, 45, 50, 59, 65, 77,
              85, 92, 101, 110, 121, 132, 144, 158, 172, 188, 206,
              225, 245, 268, 293, 320, 350]


transform_pretrain = T.Compose([
    # T.Resize(256), # Resize images to 256 x 256
    # T.CenterCrop(224), # Center crop image
    # T.RandomHorizontalFlip(),
    T.ToTensor(),  # Converting cropped images to tensors
    # T.Normalize(mean=[0.485, 0.456, 0.406],
    #            std=[0.229, 0.224, 0.225])
])
transform_test = T.Compose([
    # T.Resize(256),
    T.ToTensor(),  # Conver ting cropped images to tensors
    # T.Normalize(mean=[0.485, 0.456, 0.406],
    #            std=[0.229, 0.224, 0.225])
])

data_transforms = {'train': transform_pretrain,
                   'val': transform_test}

transform_rob = T.Compose([T.ToTensor()])

data_dir = 'tiny-imagenet-200'
image_datasets = {x: datasets.ImageFolder(os.path.join(data_dir, x), data_transforms[x])
                  for x in ['train', 'val']}

def shuffle_datasets(d1, d2):
    assert len(d1) == len(d2)
    indices = np.arange(len(d1))
    np.random.shuffle(indices)
    return torch.utils.data.Subset(d1, indices), torch.utils.data.Subset(d2, indices)

# train_loader = torch.utils.data.DataLoader(image_datasets['train'], batch_size=batch_size, shuffle=False,
#                                            num_workers=32)

# robust_train_loader = torch.utils.data.DataLoader(image_datasets['train'], batch_size=batch_size, shuffle=False,
#                                                   num_workers=32)

from layers import Linear


class Preprocess(nn.Module):
    def __init__(self, mean, std):
        super(Preprocess, self).__init__()
        self.register_buffer('mean', torch.Tensor(mean))
        self.register_buffer('std', torch.Tensor(std))

    def forward(self, input):
        input = T.functional.resize(input, 224)
        # Broadcasting
        mean = self.mean.reshape(1, 3, 1, 1)
        std = self.std.reshape(1, 3, 1, 1)
        return (input - mean) / std


def fill_gd(self, alpha):
    self[1].fill_gd(alpha)


def kill_gd(self):
    self[1].kill_gd()


nn.Sequential.fill_gd = fill_gd
nn.Sequential.kill_gd = kill_gd


def create_norm_resnet18():
    base_model = resnet18(pretrained=False)
    num_ftrs = base_model.fc.in_features
    base_model.fc = Linear(num_ftrs, 200)
    preprocess_layer = Preprocess(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    return nn.Sequential(
        preprocess_layer,
        base_model
    )


def robust_test(model, device, robust_train_loader, test_loader, criterion, num_classes, optimizer, epoch):
    train(model, criterion, device, num_classes, robust_train_loader, optimizer, epoch)


def run():
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    if loss_name == 'CrossEntropyLoss':
        criterion = nn.CrossEntropyLoss()
        criterion_summed = nn.CrossEntropyLoss(reduction='sum')


    # lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=200)

    val_loader = torch.utils.data.DataLoader(image_datasets['val'], batch_size=test_bs, shuffle=False,
                                             num_workers=32)
    cur_epochs = []
    if args.load:
        f_name = f'saved_datasets/{file_prefix}.pickle'
        print(f'Reading file: {f_name}')
        with open(f_name, 'rb') as handle:
            dataset = pickle.load(handle)
    else:
        model = create_norm_resnet18().to(device)
        model.load_state_dict(torch.load("resnet18tinyimgnet.ckpt"))

        optimizer = optim.SGD(model.parameters(),
                              lr=lr,
                              momentum=momentum,
                              weight_decay=weight_decay
                              )

        lr_scheduler = optim.lr_scheduler.MultiStepLR(optimizer,
                                                      milestones=epochs_lr_decay,
                                                      gamma=lr_decay)

        dataset = {}
        for epoch in range(1, 12):
            if epoch == 1:
                train_dataset, robust_train_dataset = shuffle_datasets(image_datasets['train'], image_datasets['train'])
                train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
                robust_train_loader = torch.utils.data.DataLoader(robust_train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
            if epoch > 1:
                robust_train_dataset = TensorDataset(robust_dataset, robust_label)
                train_dataset, robust_train_dataset = shuffle_datasets(train_dataset, robust_train_dataset)
                train_loader = torch.utils.data.DataLoader(train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
                robust_train_loader = torch.utils.data.DataLoader(robust_train_dataset, batch_size=batch_size, shuffle=False, num_workers=32)
            if epoch % 1 == 0:
                dataset[str(epoch)] = robust_train_dataset
            robust_dataset, robust_label = train_robust_data(model,
                                                             criterion, device, train_loader, robust_train_loader,
                                                             optimizer, epoch, attack_type=args.type, alpha0=alpha0, beta=beta, eps=eps,
                                                             step_size=alpha, steps=steps)
            test(model, device, val_loader)
        #     lr_scheduler.step()
        #     del model
        #     del robust_train_loader
        #     print(torch.cuda.memory_allocated())
        #     if epoch%10==0:
        #     test(model, device, C, analysis_test_loader)

        # save the dataset:

        with open(f'saved_datasets/{file_prefix}.pickle', 'wb') as handle:
            pickle.dump(dataset, handle, protocol=pickle.HIGHEST_PROTOCOL)


    print('Fine tuning:')
    # for key in dataset:
    #     print(key)
    #     robust_train_loader = torch.utils.data.DataLoader(dataset[key], batch_size=batch_size, shuffle=False)
    #     model2 = create_norm_resnet18().to(device)
    #     model2.load_state_dict(torch.load("resnet18tinyimgnet.ckpt"))
    #     optimizer = optim.SGD(model2.parameters(),
    #                           lr=0.01,
    #                           momentum=momentum,
    #                           weight_decay=weight_decay
    #                           )
    #     robust_test(model2, device, robust_train_loader, val_loader, criterion, 200, optimizer, 0)
    #     attack(model2, val_loader, device, eps=eps, alpha=alpha, steps=steps)
    #     test(model2, device, val_loader)

    if args.prop is not None:
        num_data = int(len(dataset['1']) * args.prop)
        print(f'Taking subset of dataset: {num_data}')
        indices = list(range(num_data))
        for key in dataset:
            dataset[key] = torch.utils.data.Subset(dataset[key], indices)

    for key in dataset:
        # if int(key) not in [10]:
        #     continue
        # if int(key) == 1:
        #     continue
        print("************", key, "*********************")
        robust_train_loader = torch.utils.data.DataLoader(dataset[key], batch_size=batch_size, shuffle=False)

        model22 = create_norm_resnet18().to(device)
        model22.load_state_dict(torch.load("resnet18tinyimgnet.ckpt"))
        optimizer = optim.SGD(model22.parameters(),
                              lr=0.01,
                              momentum=momentum,
                              weight_decay=weight_decay
                              )

        for i in range(21):
            # if i != 0:
            #     break
            robust_test(model22, device, robust_train_loader, val_loader, criterion, 200, optimizer, i)
            if i % 3 == 0:
                attack_type = f'auto_{args.type}' if args.auto else args.type
                print(f'attack_type: {attack_type}')
                attack(model22, val_loader, device, attack_type=attack_type, eps=eps, alpha=alpha, steps=steps)
                test(model22, device, val_loader)



def args_parser():
    parser = argparse.ArgumentParser()

    # command line arguments
    parser.add_argument('--lr', type=float, default=None)
    #parser.add_argument('--epochs', type=int, default=None)
    #parser.add_argument('--bs', type=int, default=None)
    parser.add_argument('--test_bs', type=int, default=None)

    parser.add_argument('--alpha', type=float, default=None)
    parser.add_argument('--eps', type=float, default=None)
    parser.add_argument('--beta', type=float, default=None)
    parser.add_argument('--alpha0', type=float, default=None)

    parser.add_argument('--type', type=str, default='inf')
    parser.add_argument('--auto', type=int, default=0)
    parser.add_argument('--load', type=int, default=0)
    parser.add_argument('--prop', type=float, default=None)

    return parser.parse_args()


if __name__ == "__main__":
    # Best lr after hyperparameter tuning

    args = args_parser()

    lr = 0.01 if args.lr is None else args.lr
    batch_size = 500
    test_bs = 500 if args.test_bs is None else args.test_bs

    epochs = 100
    epochs_lr_decay = [epochs / 3, epochs * 2 / 3]


    momentum = 0.9
    weight_decay = 5e-4

    eps = 2 / 255 if args.eps is None else args.eps
    alpha = 0.3 / 255 if args.alpha is None else args.alpha
    steps = 10
    beta = 0.5 / 255 if args.beta is None else args.beta
    alpha0 = 0.01 if args.alpha0 is None else args.alpha0

    file_prefix = f'lr{lr}_bs{batch_size}_eps{eps}_beta{beta}_alpha0{alpha0}_alpha{alpha}_type{args.type}'
    print(file_prefix)
    run()
